// Version information for the "groupsock" library
// Copyright (c) 1996-2010 Live Networks, Inc.  All rights reserved.

#ifndef _GROUPSOCK_VERSION_HH
#define _GROUPSOCK_VERSION_HH

#define GROUPSOCK_LIBRARY_VERSION_STRING	"2010.03.08"
#define GROUPSOCK_LIBRARY_VERSION_INT		1268006400

#endif
